"""TorchOK Model creation / weight loading / state_dict helpers.

Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/helpers.py

Copyright 2020 Ross Wightman
Licensed under The Apache 2.0 License [see LICENSE for details]
"""
from typing import Any, Callable, Dict

import torch.nn as nn
from torch.hub import load_state_dict_from_url


def build_model_with_cfg(model_cls: Callable,
                         pretrained: bool,
                         default_cfg: Dict[str, Any],
                         model_cfg: Dict[str, Any] = None,
                         **model_args) -> nn.Module:
    """Build model with specified default_cfg and optional model_args.

    Args:
        model_cls: model class
        pretrained: load pretrained weights
        default_cfg: model's default pretrained/task config
        model_cfg: Configuration for creating the model.
        **model_args: model args passed through to model __init__
    """
    model = model_cls(**model_args) if model_cfg is None else model_cls(cfg=model_cfg, **model_args)
    pretrained_url = default_cfg.get('url', None)

    if pretrained and pretrained_url is not None:
        state_dict = load_state_dict_from_url(pretrained_url, progress=False, map_location='cpu')
        model.load_state_dict(state_dict, strict=True)

    return model



from typing import Dict, Union

import torch
from omegaconf import DictConfig

from torchok.constructor import BACKBONES, HEADS, POOLINGS, TASKS
from torchok.tasks.base import BaseTask


@TASKS.register_class
class ClassificationTask(BaseTask):
    """A class for image classification task."""

    def __init__(self, hparams: DictConfig):
        """Init ClassificationTask.

        Args:
            hparams: Hyperparameters that set in yaml file.
        """
        super().__init__(hparams)
        backbones_params = self._hparams.task.params.backbone_params
        self.backbone = BACKBONES.get(self._hparams.task.params.backbone_name)(**backbones_params)

        pooling_params = self._hparams.task.params.get('pooling_params', dict())
        pooling_in_features = self.backbone.get_forward_output_channels()
        pooling_name = self._hparams.task.params.get('pooling_name', 'Identity')
        self.pooling = POOLINGS.get(pooling_name)(in_features=pooling_in_features, **pooling_params)
        
        head_params = self._hparams.task.params.get('head_params', dict())
        head_in_features = self.pooling.get_forward_output_channels()

        head_name = self._hparams.task.params.get('head_name', 'Identity')
        self.head = HEADS.get(head_name)(in_features=head_in_features, **head_params)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward method."""
        x = self.backbone(x)
        x = self.pooling(x)
        x = self.head(x)
        return x

    def forward_with_gt(self, batch: Dict[str, Union[torch.Tensor, int]]) -> Dict[str, torch.Tensor]:
        """Forward with ground truth labels."""
        input_data = batch['image']
        target = batch['target']
        freeze_backbone = self._hparams.task.params.get('freeze_backbone', False)
        with torch.set_grad_enabled(not freeze_backbone and self.training):
            features = self.backbone(input_data)
        embeddings = self.pooling(features)
        prediction = self.head(embeddings, target)
        output = {'target': target, 'embeddings': embeddings, 'prediction': prediction}
        return output



from torchok.models.backbones.resnet import (
    create_resnet,
    resnet18,
    resnet34,
    resnet50,
    resnet101,
    resnet152,
    seresnet18,
    seresnet34,
    seresnet50,
    seresnet101,
    seresnet152,
)
from torchok.models.backbones.hrnet import (
    create_hrnet,
    hrnet_w18_small,
    hrnet_w18_small_v2,
    hrnet_w18,
    hrnet_w30,
    hrnet_w32,
    hrnet_w40,
    hrnet_w44,
    hrnet_w48,
    hrnet_w64,
)

__all__ = [
    'ResNet',
    'create_resnet',
    'resnet18',
    'resnet34',
    'resnet50',
    'resnet101',
    'resnet152',
    'seresnet18',
    'seresnet34',
    'seresnet50',
    'seresnet101',
    'seresnet152',
    'create_hrnet',
    'hrnet_w18_small',
    'hrnet_w18_small_v2',
    'hrnet_w18',
    'hrnet_w30',
    'hrnet_w32',
    'hrnet_w40',
    'hrnet_w44',
    'hrnet_w48',
    'hrnet_w64',
]



"""
This HRNet implementation is modified from the following repository:
https://github.com/HRNet/HRNet-Semantic-Segmentation
"""
import torch.nn as nn
from torch import Tensor

from torchok.constructor import HEADS
from torchok.models.heads.base import AbstractHead
from torchok.models.modules.bricks.convbnact import ConvBnAct


@HEADS.register_class
class HRNetSegmentationHead(AbstractHead):
    """HRNet head for segmentation tasks."""
    def __init__(self, in_features: int, num_classes: int):
        """Init HRNetSegmentationHead.

        Args:
            in_features: Size of each input sample.
            num_classes: Number of classes.
        """
        super().__init__(in_features, num_classes)
        self.num_classes = num_classes
        self.convbnact = ConvBnAct(in_features,
                                   in_features,
                                   kernel_size=1,
                                   padding=0,
                                   stride=1)
        self.final_conv_layer = nn.Conv2d(in_channels=in_features,
                                          out_channels=num_classes,
                                          kernel_size=1,
                                          stride=1,
                                          padding=0)

    def forward(self, x: Tensor) -> Tensor:
        """Forward method"""
        x = self.convbnact(x)
        x = self.final_conv_layer(x)
        return x



from typing import Optional

import torch.nn as nn
from torch import Tensor

from torchok.models.modules.bricks.convbnact import ConvBnAct


class Bottleneck(nn.Module):
    """Bottleneck building block for ResNet architecture."""

    expansion = 4

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 stride: int = 1,
                 downsample: Optional[nn.Module] = None,
                 attn_layer: Optional[nn.Module] = None):
        """Init Bottleneck.

        Args:
            in_channels: Number of input channels.
            out_channels: Number of output channels.
            stride: Stride.
            downsample: Downsample module.
            attn_layer: Attention block.
        """
        super().__init__()
        out_block_channels = out_channels * self.expansion

        self.convbnact1 = ConvBnAct(in_channels, out_channels, kernel_size=1, padding=0, stride=1)
        self.convbnact2 = ConvBnAct(out_channels, out_channels, kernel_size=3, padding=1, stride=stride)
        self.convbnact3 = ConvBnAct(out_channels, out_block_channels, kernel_size=1, padding=0, act_layer=None)
        self.act = nn.ReLU(inplace=True)
        self.se = attn_layer(out_block_channels) if attn_layer is not None else None
        self.downsample = downsample

    def forward(self, x: Tensor) -> Tensor:
        """Forward method."""
        identity = x

        out = self.convbnact1(x)
        out = self.convbnact2(out)
        out = self.convbnact3(out)

        if self.se is not None:
            out = self.se(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.act(out)

        return out



from typing import List, Union

from torch import Tensor
from abc import ABC, abstractmethod
from torchok.models.base import BaseModel


class AbstractHead(BaseModel, ABC):
    """An abstract class for head."""

    def __init__(self, in_features, out_features):
        """Init AbstractHead.

        Args:
            in_features: Input features.
            out_features: Output features.
        """
        super().__init__()
        self._in_features = in_features
        self._out_features = out_features

    @abstractmethod
    def forward(self, *args, **kwargs) -> Tensor:
        """Forward method."""
        pass

    def get_forward_output_channels(self) -> Union[int, List[int]]:
        """Return number of output channels."""
        return self._out_features

    @property
    def in_features(self) -> int:
        """Input features."""
        return self._in_features

    @property
    def out_features(self) -> int:
        """Output features."""
        return self._out_features



from typing import List, Union

import torch
from torch import Tensor

import torch.nn.functional as F
from torchok.constructor import NECKS
from torchok.models.base import BaseModel
from torchok.models.modules.bricks.convbnact import ConvBnAct


@NECKS.register_class
class HRNetSegmentationNeck(BaseModel):
    """HRNet neck for segmentation task. """
    def __init__(self, in_channels):
        """Init HRNetSegmentationNeck.

        Args:
            in_channels: Input channels.
        """
        super().__init__()
        self.in_channels = sum(in_channels)

    def forward(self, features: List[Tensor]) -> Tensor:
        """Forward method."""
        input_image, *features = features
        interpolated_feat = []
        for feature in features:
            interpolated = F.interpolate(feature, size=input_image.shape[2:], mode='bilinear', align_corners=True)
            interpolated_feat.append(interpolated)

        feats = torch.cat(interpolated_feat, 1)
        return feats

    def get_forward_output_channels(self) -> Union[int, List[int]]:
        """Return number of output channels."""
        return self.in_channels



from abc import ABC, abstractmethod

from sklearn.preprocessing import normalize
from typing import Callable, List, Optional, Union, Dict, Tuple, Generator

import torch
import faiss
import pandas as pd
import numpy as np
import math

from torchmetrics import Metric
from ranx.metrics import precision, recall, average_precision, ndcg
from enum import Enum

from torchok.constructor import METRICS


__all__ = [
    'PrecisionAtKMeter',
    'RecallAtKMeter',
    'MeanAveragePrecisionAtKMeter',
    'NDCGAtKMeter',
]


class DatasetType(Enum):
    CLASSIFICATION = 'classification'
    REPRESENTATION = 'representation'


class MetricDistance(Enum):
    IP = 'IP'
    L2 = 'L2'


dataset_enum_mapping = {
    'classification': DatasetType.CLASSIFICATION,
    'representation': DatasetType.REPRESENTATION,
}

distance_enum_mapping = {
    'IP': MetricDistance.IP,
    'L2': MetricDistance.L2
}


class IndexBasedMeter(Metric, ABC):
    """Base class for representation metrics.

    Store retrieval vectors and targets during phase in update method. FAISS library is used to build an index 
    and search for top-k in it. Supports 2 datasets: classification dataset with targets, 
    and representation dataset with scores and queries_idxs tensors.
    Compute method return generator with relevant and closest (FAISS searched) indexes. The relevant index
    contain it's relevant index with scores for current query. And the closest contain closest index with it's distance.
    """
    def __init__(self, exact_index: bool, dataset_type: str, metric_distance: str, \
        metric_func: Callable, k: Optional[int] = None, search_batch_size: Optional[int] = None, \
        normalize_vectors: bool = False, **kwargs):
        """Initialize IndexBasedMeter.

        Args:
            exact_index: If true then build fair inner product or Euclidean index (depends on the metric chosen), 
                otherwise the index will be an approximate nearest neighbours search index.
            dataset_type: Dataset type (classification or representation), which will be used to calculate metric.
            metric_distance: Metric distance (IP - cosine distance, L2 - euclidean), which will be used to build 
                FAISS index.
            metric_func: Representation metric (e.g ranx metric function) with the follow backend 
                `def metric_func(
                    qrels: Union[np.ndarray, numba.typed.List],
                    run: Union[np.ndarray, numba.typed.List],
                    k: int = 0,
                ) -> np.ndarray`
                where qrels - y_true and run - y_pred
                see https://github.com/AmenRa/ranx/blob/ccab1549de81e7366e34213c86089e965db55f72/ranx/metrics.py
                for more details. 
            k: Number of top closest indexes to get.
            search_batch_size: The size for one FAISS search request.
            normalize_vectors: If true vectors will be normalize, overwise no.

        Raises:
            ValueError: If metric or dataset is not correct write.
        """
        super().__init__(compute_on_step=False, **kwargs)
        self.exact_index = exact_index
        
        self.dataset_type = dataset_enum_mapping[dataset_type]
        self.metric_distance = distance_enum_mapping[metric_distance]
        self.metric_func = metric_func
        self.normalize_vectors = normalize_vectors
        # set search_batch_size as num CPUs if search_batch_size is None
        self.search_batch_size = torch.get_num_threads() if search_batch_size is None else search_batch_size
        self.k = k

        self.add_state('vectors', default=[], dist_reduce_fx=None)
        if self.dataset_type == DatasetType.CLASSIFICATION:
            # if classification dataset
            self.add_state('targets', default=[], dist_reduce_fx=None)
        else:
            # if representation dataset
            self.add_state('queries_idxs', default=[], dist_reduce_fx=None)
            self.add_state('scores', default=[], dist_reduce_fx=None)

    def update(self, vectors: torch.Tensor, targets: Optional[torch.Tensor] = None, \
               queries_idxs: Optional[torch.Tensor] = None, scores: Optional[torch.Tensor] = None):
        """Append tensors in storage.
        
        Args:
            vectors: Often it would be embeddings, size (batch_size, embedding_size).
            targets: The labels for every vectors in classification mode, size (batch_size).
            queries_idxs: Integer tensor where values >= 0 represent indices of queries with corresponding vectors 
                in vectors tensor and value -1 indicates that the corresponding vector isn't a query.
            scores: The scores tensor, see representation dataset for more information, 
                size (batch_size, total_num_queries).

        Raises:
            ValueError: If dataset is of classification type and targets is None, or if dataset is of representation 
                type and at least one of scores or queries_idxs is None.
        """
        vectors = vectors.detach().cpu()
        self.vectors.append(vectors)
        if self.dataset_type == DatasetType.CLASSIFICATION:
            if targets is None:
                raise ValueError("In classification dataset target must be not None.")
            targets = targets.detach().cpu()
            self.targets.append(targets)
        else:
            if queries_idxs is None:
                raise ValueError("In representation dataset queries_idxs must be not None.")
            if scores is None:
                raise ValueError("In representation dataset scores must be not None")
            
            queries_idxs = queries_idxs.detach().cpu()
            self.queries_idxs.append(queries_idxs)
            
            scores = scores.detach().cpu()
            self.scores.append(scores)

    def compute(self) -> float:
        """Compute metric value.
        
        Firstly it gathers all tensors in storage (done by torchmetrics). 
        Then it prepares data, separates query and database vectors. 
        Then it builds the FAISS index. 
        Then, it create a generator of relevant and closest arrays.
        Finally, it compute metric.

        Returns:
            Metric value.
        """
        vectors = torch.cat(self.vectors).numpy()
        
        if self.normalize_vectors:
            vectors = normalize(vectors)

        if self.dataset_type == DatasetType.CLASSIFICATION:
            # if classification dataset
            targets = torch.cat(self.targets).numpy()
            # prepare data
            q_vecs, db_vecs, relevants, scores, \
                db_idxs, q_order_idxs = self.prepare_classification_data(vectors, targets)
        else:
            # if representation dataset
            scores = torch.cat(self.scores).numpy()
            queries_idxs = torch.cat(self.queries_idxs).numpy()
            # prepare data
            q_vecs, db_vecs, relevants, scores, \
                db_idxs, q_order_idxs = self.prepare_representation_data(vectors, queries_idxs, scores)

        q_vecs = q_vecs.astype(np.float32)
        db_vecs = db_vecs.astype(np.float32)

        index = self.build_index(db_vecs)

        # if k is None set it as database length
        k = len(db_vecs) if self.k is None else self.k

        # create relevant, closest generator
        generator = self.query_generator(index, relevants, q_vecs, db_idxs, q_order_idxs, k, scores)
        
        # compute metric
        scores = []
        for relevant_idx, closest_idx in generator:
            scores += self.metric_func(relevant_idx, closest_idx, k=k).tolist()
        return np.mean(scores)

    def prepare_representation_data(self, vectors: np.ndarray, queries_idxs: np.ndarray, scores: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Prepare data for faiss build index, and following search, in representation dataset case.
        
        Separate query and database vectors from storage vectors.
        Prepare scores.
        Generate relevant indexes for every query request.

        Args:
            vectors: Vectors of the whole storage, includes queries and database vectors, 
                size (dataset_size, embedding_size).
            queries_idxs: Integer tensor where values >= 0 represent indices of queries with corresponding vectors 
                in vectors tensor and value -1 indicates that the corresponding vector isn't a query.
            scores: The scores tensor, see representation dataset for more information, 
                size (batch_size, total_num_queries).

        Returns:
            q_vecs: Queries vectors, size (queries_size, embedding_size).
            db_vecs: Database vectors, for faiss build index, size (database_size, embedding_size).
            relevant: Array of arrays relevant indexes in database for every query request, size (queries_size, ).
            scores: Array of scores related to queries that have at least one relevant item.
            db_idxs: Array with all database indexes.
            queries_idxs: Array of queries order number.
        """
        is_queries = queries_idxs >= 0
        queries_idxs = queries_idxs[is_queries]
        q_vecs = vectors[is_queries]
        db_vecs = vectors[~is_queries]
        db_idxs = np.arange(len(db_vecs))
        scores = scores[~is_queries]
        
        relevant = []
        empty_relevant_idxs = []

        for idx in range(len(q_vecs)):
            relevant_idxs = np.where(scores[:, queries_idxs[idx]] > 0.)[0]
            if len(relevant_idxs) == 0:
                empty_relevant_idxs.append(idx)
            else:
                # Need to sort relevant indexes by its scores for NDCG metric
                current_scores = scores[relevant_idxs, queries_idxs[idx]]
                sort_indexes = np.argsort(current_scores)
                relevant_idxs = relevant_idxs[sort_indexes[::-1]]
                relevant.append(relevant_idxs)
        
        relevant = np.array(relevant)

        # remove empty relevant queries
        q_vecs = np.delete(q_vecs, empty_relevant_idxs, axis=0)
        queries_idxs = np.delete(queries_idxs, empty_relevant_idxs)
        
        return q_vecs, db_vecs, relevant, scores, db_idxs, queries_idxs

    def prepare_classification_data(self, vectors: np.ndarray, targets: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Prepare data for faiss build index, and following search, in classification dataset case.
        
        Separate query and database vectors from storage vectors.
        Query vector index will be the first row index for each unique target value found in targets array. 
        So, total number of queries is equal to unique target count. 
        The remaining vectors will be treated as database vectors (again for each unique target value)

        Args:
            vectors: Vectors of the whole storage, includes queries and database vectors, 
                size (dataset_size, embedding_size).
            targets: Targets in classification task for every vector, size (database_size).

        Retruns:
            q_vecs: Queries vectors, size (queries_size, embedding_size).
            db_vecs: Database vectors, for faiss build index, size (database_size, embedding_size).
            relevant: Array of relevant indexes in database for every query request, size (queries_size, ).
            scores: Array of scores without queries empty scores.
            db_idxs: Array with all database indexes.
            queries_idxs: Array of queries order number.
        """
        ts = pd.Series(targets)

        # Group item indices of the same class
        groups = pd.Series(ts.groupby(ts).groups)

        # Take the first index of the group as a query
        # If there is only one element in the group, ignore this group
        query_idxs = groups.apply(lambda x: x[0] if len(x) > 1 else -1)

        # Rest indices are used as a relevant/database elements
        relevant = groups.apply(lambda x: x[1:].values if len(x) > 1 else x.values)

        # Create mapping from database index to original index
        db_idxs = np.array(sorted([i for j in relevant for i in j]))

        # Filter groups with more than one sample per group
        correct_classes = query_idxs != -1
        relevant = relevant[correct_classes]
        query_idxs = query_idxs[correct_classes]

        # Retrieve query and database vectors and build index base on latter.
        q_vecs = vectors[query_idxs]
        db_vecs = vectors[db_idxs]

        scores = None
        query_order_idxs = np.arange(len(q_vecs))
        return q_vecs, db_vecs, relevant, scores, db_idxs, query_order_idxs

    def query_generator(self, index: Union[faiss.swigfaiss_avx2.IndexFlatIP, faiss.swigfaiss_avx2.IndexFlatL2], \
                        relevants: np.ndarray, queries: np.ndarray, db_ids: np.ndarray, q_order_idxs: np.ndarray, \
                        k: int, scores: Optional[np.ndarray] = None
    ) -> Generator[Tuple[List[np.ndarray], List[np.ndarray]], None, None]:
        """Create relevants relevant, closest arrays.

        Output in relevant array, contain it's index in database and score for current query.
        Output in closest array, contain it's index in database and distance for current query.
        
        Args:
            index: Faiss database built index.
            relevants: Relevant indexes for every query, size (total_num_queries, ) and the second shape is can be different
                for every query request.
            queries: Vectors for every query request, size (total_num_queries, embedding_size).
            scores: Scores for every relevant index per each query request, size (database_size, total_num_queries).
                See representation dataset for more information.
            db_ids: Database indexes.
            k:  Number of top closest indexes to get.

        Returns:
            Generator wich contain relevant and closest Tuple values. 

            Relevant include relevant indexes and scores, size (search_batch_size, , 2).
            Closest include searched indexes and distances, size (search_batch_size, , 2).
        """
        def generator():
            """Generate relevant - y_true, and closest - y_pred for metric calculation with ranx library.

            Returns:
                relevant: List of relevant indexes with its scores per each queries. Length of list = search_batch_size.
                    And size of one element of list = (relevant_indexes_size, 2), where last shape 2 for relevant index
                    and it score. 
                    Example for 3 search_batch_size, and relevant_sizes = [2, 2, 1] with score = 1 for every \
                    relevant index:
                    [
                        np.array([[6, 1], [7, 1]]), 
                        np.array([[2, 1], [5, 1]]), 
                        np.array([[4, 1]])
                    ].
                closest: List of numpy arrays, with nearest searched indexes by top k.
                    Example for k = 3:
                    [
                        np.array([[4, 1],
                               [2, 1],
                               [6, 1]]), 
                        np.array([[5, 1],
                               [2, 1],
                               [6, 1]]), 
                        np.array([[4, 1],
                               [5, 1],
                               [6, 1]])
                    ].
            """
            for i in range(0, len(queries), self.search_batch_size):
                query_idxs = np.arange(i, min(i + self.search_batch_size, len(queries)))

                closest_dist, closest_idx = index.search(queries[query_idxs], k=k)
                relevant = relevants[query_idxs]
 
                if self.metric_distance == MetricDistance.IP:
                    closest_dist = 1 - closest_dist

                # NDCG score=distance is needed to sort more relevant examples, but in this part of code we had 
                # already sorted our examples by faiss. So if we change score = 1 to distance with type float 
                # the index of relevant will be also float and after that inside ranx it may be fail to compare 
                # relevant int index with our relevant float index.
                closest = map(lambda idx: np.stack((db_ids[closest_idx[idx]], [1] * len(closest_idx[idx])), axis=1), \
                    np.arange(len(closest_idx)))
                
                if scores is None:
                    relevant = map(lambda r: np.stack((r, np.ones_like(r)), axis=1), relevant)
                else:
                    relevant = map(lambda r_q: np.stack((r_q[0], scores[r_q[0], q_order_idxs[r_q[1]]]), axis=1), \
                        zip(relevant, query_idxs))
                
                relevant = list(relevant)
                closest = list(closest)

                yield relevant, closest

        return generator()

    def build_index(self, vectors: np.ndarray):
        """Build index of a given set of vectors with FAISS.

        Args:
            vectors: Database vectors to index, size (database_size, embedding_size).

        Returns:
            Constructed index.
        """
        vectors = vectors.astype(np.float32)
        n, d = vectors.shape

        index_class = faiss.IndexFlatIP if self.metric_distance == MetricDistance.IP else faiss.IndexFlatL2
        if self.exact_index:
            index = index_class(d)
        else:
            nlist = 4 * math.ceil(n ** 0.5)
            quantiser = index_class(d)
            index = faiss.IndexIVFFlat(quantiser, d, nlist, self.metric_distance.value)
            index.train(vectors)

        index.add(vectors)
        return index


@METRICS.register_class
class PrecisionAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True, \
                 metric_distance: str = 'IP', k: Optional[int] = None, \
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance, \
            metric_func=precision, k=k, search_batch_size=search_batch_size, normalize_vectors=normalize_vectors, \
            **kwargs)


@METRICS.register_class
class RecallAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True, \
                 metric_distance: str = 'IP', k: Optional[int] = None, \
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance, \
            metric_func=recall, k=k, search_batch_size=search_batch_size, normalize_vectors=normalize_vectors, \
            **kwargs)


@METRICS.register_class
class MeanAveragePrecisionAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True, \
                 metric_distance: str = 'IP', k: Optional[int] = None, \
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance, \
            metric_func=average_precision, k=k, search_batch_size=search_batch_size, \
            normalize_vectors=normalize_vectors, **kwargs)


@METRICS.register_class
class NDCGAtKMeter(IndexBasedMeter):
    def __init__(self, dataset_type: str, exact_index: bool = True, \
                 metric_distance: str = 'IP', k: Optional[int] = None, \
                 search_batch_size: Optional[int] = None, normalize_vectors: bool = False, **kwargs):
        super().__init__(exact_index=exact_index, dataset_type=dataset_type, metric_distance=metric_distance, \
            metric_func=ndcg, k=k, search_batch_size=search_batch_size, normalize_vectors=normalize_vectors, \
            **kwargs)
    


import numbers
import numpy as np

from torch import Tensor, tensor, nn
from torchmetrics import Metric

from dataclasses import dataclass, field
from enum import Enum
from typing import List, Dict, Any, Optional

from torchok.constructor import METRICS
from torchok.constructor.config_structure import MetricParams, Phase


class MetricWithUtils(nn.Module):
    """Union class for metric and metric utils parameters."""
    def __init__(self, metric: Metric, mapping: Dict[str, str], log_name: str):
        """Initalize MetricWithUtils.
        
        Args:
            metric: Metric written with TorchMetrics.
            mapping: Dictionary for mapping Metric forward input keys with Task output dictionary keys.
            log_name: The metric name used in logs.
        """
        super().__init__()
        self.__metric = metric
        self.__mapping = mapping
        self.__log_name = log_name

    @property
    def log_name(self) -> str:
        """The metric name used in logs."""
        return self.__log_name

    @property
    def mapping(self) -> Dict[str, str]:
        """Dictionary for mapping Metric forward input keys with Task output dictionary keys."""
        return self.__mapping

    def forward(self, *args, **kwargs):
        return self.__metric(*args, **kwargs)

    def compute(self):
        return self.__metric.compute()


class MetricsManager(nn.Module):
    """Manages all metrics for a Task."""
    def __init__(self, params: List[MetricParams]):
        """Initialize MetricManager.

        Args:
            params: Metric parameters.
        """
        super().__init__()
        self.__phase2metrics = nn.ModuleDict()
        for phase in Phase:
            self.__phase2metrics[phase.name] = self.__get_phase_metrics(params, phase) 

    def __get_phase_metrics(self, params: List[MetricParams], phase: Phase) -> nn.ModuleList:
        """Generate metric list for current phase.

        Args:
            params: All metric params from config file.
            phase: Current phase Enum.

        Returns:
            metrics: Metric list as nn.ModuleList for current phase.

        Raises:
            ValueError: If got two identical log_names.
        """
        # create added_metric_names list
        added_log_names = []
        metrics = []
        for metric_params in params:
            if phase not in metric_params.phases:
                continue
            metric = METRICS.get(metric_params.name)(**metric_params.params)
            mapping = metric_params.mapping
            prefix = '' if metric_params.prefix is None else metric_params.prefix + '_'
            log_name = prefix + metric_params.name
            if log_name in added_log_names:
                raise ValueError(f'Got two metrics with identical names: {log_name}. '
                                 f'Please, set differet prefixes for identical metrics in the config file.')           
            else:
                added_log_names.append(log_name)

            metrics.append(MetricWithUtils(metric=metric, mapping=mapping, log_name=log_name))

        metrics = nn.ModuleList(metrics)

        return metrics

    def forward(self, phase: Phase, *args, **kwargs):
        """Update states of all metrics on phase loop.

        Args:
            phase: Phase Enum.
        """
        args = list(args)

        for metric_with_utils in self.__phase2metrics[phase.name]:
            targeted_kwargs = self.map_arguments(metric_with_utils.mapping, kwargs)
            metric_with_utils(*args, **targeted_kwargs)
            
    def on_epoch_end(self, phase: Phase) -> Dict[str, Tensor]:
        """Summarize epoch values and return log.
        
        Args:
            phase: Run metric phase.

        Returns:
            log: Logging dictionary, where key is `<phase>/<metric_name>` and value is metric value for
            a given phase.

        Raises:
            ValueError: If metric.compute() returns not numerical value.
        """ 
        log = {}
        for metric_with_utils in self.__phase2metrics[phase.name]:
            metric_value = metric_with_utils.compute()
            # If it tensor type with wrong shape.
            if isinstance(metric_value, Tensor) and len(metric_value.shape) != 0:
                raise ValueError(f'{metric_with_utils.log_name} must compute number value, ' 
                                 f'not torch tensor with shape {metric_value.shape}.')
            # If it numpy array with wrong shape.
            if isinstance(metric_value, np.ndarray) and len(metric_value.shape) != 0:
                raise ValueError(f'{metric_with_utils.log_name} must compute number value, ' 
                                 f'not numpy array with shape {metric_value.shape}.')
            # If it numpy array with one element but wrong dtype
            if (isinstance(metric_value, np.ndarray) and len(metric_value.shape) == 0 and 
                np.issubdtype(metric_value.dtype, np.number)):
                raise ValueError(f'{metric_with_utils.log_name} must compute number value, ' 
                                 f'not numpy array element with dtype {metric_value.dtype}.')

            is_number = isinstance(metric_value, numbers.Number)
            # If not numeric type.
            if not (is_number or isinstance(metric_value, Tensor) or  isinstance(metric_value, np.ndarray)):
                raise ValueError(f'{metric_with_utils.log_name} must compute number value, ' 
                                 f'not numpy array element with dtype {metric_value.dtype}.')
            
            metric_key = f'{phase.value}/{metric_with_utils.log_name}'
            log[metric_key] = metric_value

        return log

    @staticmethod
    def map_arguments(mapping: Dict[str, str], task_output: Dict[str, Any]) -> Dict[str, Any]:
        # TODO: create a common function for MetricManager and Constructor
        """Map arguments between metric target_fields and task output dictionary.

        Args:
            mapping: Dictionary for mapping Metric forward input keys with Task output dictionary keys.
            task_output: Output after task forward pass.

        Returns:
            metric_input: Metric input dictionary like ``**kwargs`` for metric forward pass.

        Raises:
            ValueError: If not found mapping_source in task_output keys.
        """
        metric_input = {}
        for metric_target, metric_source in mapping.items():
            if metric_source in task_output:
                arg = task_output[metric_source]
                metric_input[metric_target] = arg
            else:
                raise ValueError(f'Cannot find {metric_source} for your mapping {metric_target} : {metric_source}. '
                                 f'You should either add {metric_source} output to your model or remove the mapping '
                                 f'from configuration')
        return metric_input

    @property
    def phase2metrics(self) -> Dict[Phase, nn.ModuleList]:
        """Dictionary of phase to their metrics list with type nn.ModuleList([MetricWithUtils])"""
        return self.__phase2metrics



from torchok.models.base import (
    BaseModel,
    FeatureHooks,
    FeatureInfo,
    HookType
)
from torchok.models import backbones
from torchok.models import identity
from torchok.models import heads
from torchok.models import necks
from torchok.models import poolings

from torchok.constructor import HEADS, POOLINGS


HEADS.register_class(identity.Identity)
POOLINGS.register_class(identity.Identity)


__all__ = [
    'BaseModel',
    'FeatureHooks',
    'FeatureInfo',
    'HookType',
]



"""TorchOK Squeeze-and-Excitation Channel Attention module.

Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/models/layers/squeeze_excite.py
Copyright 2019 Ross Wightman
Licensed under The Apache 2.0 License [see LICENSE for details]
"""
from typing import Optional, Union

from torch import nn as nn


class SEModule(nn.Module):
    """SE Module as defined in original SE-Nets with a few additions."""

    def __init__(self,
                 channels: int,
                 reduction: int = 16,
                 reduction_ratio: Optional[float] = None,
                 reduction_channels: Optional[int] = None,
                 min_channels: int = 8,
                 divisor: int = 1):
        """Init SEModule.

        Args:
            channels: Number input channels.
            reduction: Reduction coefficient.
            reduction_ratio: Reduction ratio.
            reduction_channels: Reduction channels.
            min_channels: Minimum number of reduction channels.
            divisor: `reduction_channels` must be a multiple of `divisor`.
        """
        super().__init__()

        if reduction_channels is not None:
            reduction_channels = reduction_channels
        elif reduction_ratio is not None:
            reduction_channels = self.__make_divisible(channels * reduction_ratio, divisor, min_channels)
        else:
            reduction_channels = self.__make_divisible(channels // reduction, divisor, min_channels)
        self.fc1 = nn.Conv2d(channels, reduction_channels, kernel_size=1, bias=True)
        self.act = nn.ReLU(inplace=True)
        self.fc2 = nn.Conv2d(reduction_channels, channels, kernel_size=1, bias=True)
        self.gate = nn.Sigmoid()

    def __make_divisible(self, value: Union[int, float], divisor: int = 8, min_value: Optional[int] = None):
        """Make divisible function.

        This function rounds the channel number to the nearest value that can be divisible by the divisor.

        Args:
            value: The original channel number.
            divisor: The divisor to fully divide the channel number.
            min_value: The minimum value of the output channel.
        """
        min_value = min_value or divisor
        new_value = max(min_value, int(value + divisor / 2) // divisor * divisor)
        if new_value < 0.9 * value:
            new_value += divisor
        return new_value

    def forward(self, x):
        """Forward method."""
        x_se = x.mean((2, 3), keepdim=True)
        x_se = self.fc1(x_se)
        x_se = self.act(x_se)
        x_se = self.fc2(x_se)
        return x * self.gate(x_se)



import datetime
from pathlib import Path

from pytorch_lightning import Trainer
from pytorch_lightning.loggers.tensorboard import TensorBoardLogger

from torchok.callbacks.model_checkpoint_with_onnx import ModelCheckpointWithOnnx


def create_outputs_path(log_dir, experiment_name):
    version = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    log_dir = Path(log_dir)
    full_outputs_path = log_dir / experiment_name / version
    full_outputs_path.mkdir(exist_ok=True, parents=True)

    return log_dir, experiment_name, version, full_outputs_path


def create_trainer(train_config):
    outputs_path, experiment_name, version, full_outputs_path = create_outputs_path(train_config.log_dir,
                                                                                    train_config.experiment_name)
    logger = TensorBoardLogger(outputs_path, experiment_name, version)
    checkpoint_callback = ModelCheckpointWithOnnx(**train_config.checkpoint, dirpath=str(full_outputs_path))
    callbacks = [checkpoint_callback]
    trainer = Trainer(logger=logger, callbacks=callbacks, **train_config.trainer)
    return trainer



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torchok',
 'torchok.callbacks',
 'torchok.constructor',
 'torchok.data',
 'torchok.data.datasets',
 'torchok.data.datasets.classification',
 'torchok.data.datasets.detection',
 'torchok.data.datasets.examples',
 'torchok.data.datasets.representation',
 'torchok.data.datasets.segmentation',
 'torchok.data.transforms',
 'torchok.losses',
 'torchok.losses.common',
 'torchok.losses.representation',
 'torchok.losses.segmentation',
 'torchok.metrics',
 'torchok.models',
 'torchok.models.backbones',
 'torchok.models.backbones.utils',
 'torchok.models.heads',
 'torchok.models.heads.classification',
 'torchok.models.heads.detection',
 'torchok.models.heads.representation',
 'torchok.models.heads.segmentation',
 'torchok.models.modules',
 'torchok.models.modules.blocks',
 'torchok.models.modules.bricks',
 'torchok.models.necks',
 'torchok.models.necks.classification',
 'torchok.models.necks.detection',
 'torchok.models.necks.segmentation',
 'torchok.models.poolings',
 'torchok.models.poolings.classification',
 'torchok.models.poolings.representation',
 'torchok.optim',
 'torchok.optim.optimizers',
 'torchok.optim.schedulers',
 'torchok.tasks']

package_data = \
{'': ['*']}

install_requires = \
['albumentations>=1.2,<1.3',
 'boto3>=1.24,<1.25',
 'faiss>=1.5,<1.6',
 'hydra-core>=1.2,<1.3',
 'mlflow>=1.22,<1.23',
 'numpy>=1.22,<1.23',
 'onnx>=1.12,<1.13',
 'onnxruntime-gpu>=1.11,<1.12',
 'opencv-python>=4.6,<4.7',
 'pandas>=1.4,<1.5',
 'pillow>=9.1,<9.2',
 'pytorch-lightning>=1.6,<1.7',
 'ranx>=0.2,<0.3',
 'torchmetrics>=0.8,<0.9']

setup_kwargs = {
    'name': 'torchok',
    'version': '0.4.0',
    'description': 'The toolkit for fast Deep Learning experiments in Computer Vision',
    'long_description': None,
    'author': 'Vlad Vinogradov',
    'author_email': 'vladt9@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod
import torch
import torch.nn as nn

from dataclasses import dataclass
from functools import partial
from collections import OrderedDict, defaultdict
from copy import deepcopy
from typing import Dict, List, Tuple, Union, Generator, Optional, Any
from enum import Enum


class HookType(Enum):
    FORWARD = 'forward'
    FORWARD_PRE = 'forward_pre'


@dataclass
class FeatureInfo:
    """Gather feature hooks params.
    
    Args:
        module_name: Name of hooked module.
        num_channels: Number of image channels.
        stride: Downscale value for get image shape in current module.
        hook_type: Hook type (forward hook or forward pre hook).
    """
    module_name: str
    num_channels: int
    stride: int
    hook_type: HookType = HookType.FORWARD


class FeatureHooks:
    """ Feature Hook Helper.

    This module helps with the setup and extraction of hooks for extracting features from
    internal nodes in a model by node name. This works quite well in eager Python but needs
    redesign for torcscript.
    """
    def __init__(self, features_info: List[FeatureInfo], named_modules: Generator[Tuple[str, nn.Module], None, None]):
        """Initialize feature hooks.
        
        Args:
            features_info: Hooks features info that must be registered.
            named_modules: Result of self.named_modules() function call.
        """
        # setup feature hooks
        modules = {k: v for k, v in named_modules}
        for feature_info in features_info:
            hook_name = feature_info.module_name
            module = modules[hook_name]
            hook_fn = partial(self._collect_hook_features, hook_name)
            # register hooks
            if feature_info.hook_type == HookType.FORWARD_PRE:
                module.register_forward_pre_hook(hook_fn)
            else:
                module.register_forward_hook(hook_fn)
           
        self._feature_outputs = defaultdict(OrderedDict)

    def _collect_hook_features(self, hook_name: str, *args):
        """Collect hooks features in self._feature_outputs.
        
        Args:
            hook_name: Name of saved hooked module.
        """
        # tensor we want is last argument, output for fwd, input for fwd_pre
        x = args[-1]  
        if isinstance(x, (tuple, list)) and len(x) == 1:
            # unwrap input tuple
            x = x[0]  
            device = x.device
        else:
            device = x[0].device
        self._feature_outputs[device][hook_name] = x

    def get_features(self, device: torch.device) -> Dict[str, torch.tensor]:
        """Return hooks features.

        Args:
            device: Saved hooks device.

        Returns:
            output: Dict of module hook name to its output tensor.
        """
        output = self._feature_outputs[device]
        # clear after reading
        self._feature_outputs[device] = OrderedDict()  
        return output


class BaseModel(nn.Module, ABC):
    """Base model for all TorchOk Models - Backbone, Neck, Pooling and Head.

    This class supports adding feature hooks to some model layers.
    Class has feature info to describe hooks.
    To create hooks, method `self.get_features_info()` must be overridden, and the function `self.create_hooks()` 
    must be called in the end line of initialization method of inherited class!
    Method `self.create_hooks(*args, **kwargs)` called `self.get_features_info(*args, **kwargs)` with 
    the same parameters! 
    Also class contains method get_output_channels which returns forward pass and hooks output channels.

    Example:
        >>> class ModelWithHooks(BaseModel):
        >>>     def __init__(self, input_channel: int = 3):
        >>>         super().__init__()
        >>>         self.out_features = [5, 10, 15]
        >>>         self.conv1 = nn.Conv2d(input_channel, 5, 3)
        >>>         self.conv2 = nn.Conv2d(5, 10, 3)
        >>>         self.conv3 = nn.Conv2d(10, 15, 3)
        >>>         # Call create_hooks after all modules are initialized, with parameters as in get_features_info.
        >>>         self.create_hooks(output_channels=[10, 15], module_names=['conv2', 'conv3'])
        >>>
        >>>     def get_features_info(self, output_channels, module_names):
        >>>         # Overwrite get_features_info.
        >>>         features_info = []
        >>>         for num_channels, module_name in zip(output_channels, module_names):
        >>>             feature_info = FeatureInfo(module_name=module_name, num_channels=num_channels, stride=1)
        >>>             features_info.append(feature_info)
        >>>         return features_info
        >>>
        >>>     def get_forward_output_channels(self):
        >>>         return self.out_features[-1]
        >>>
        >>> model = ModelWithHooks()
        >>> print(model.get_output_channels())
        (15, [10, 15])
    """
    def __init__(self):
        """Initialize BaseModel.
        
        Set features_info and feature_hooks as None.
        """
        super().__init__()
        self.__features_info = None
        self.__feature_hooks = None

    def get_features_info(self, *args: Any, **kwargs: Any) -> List[FeatureInfo]:
        """Initialize feature hooks.
        
        It should be overriden in an inherited class if a user is expected to have access to feature hooks 
        of the implemented Module.

        Retruns: Hooks FeatureInfo list.
        """
        return []

    def create_hooks(self, *args: Any, **kwargs: Any) -> Tuple[List[FeatureInfo], FeatureHooks]:
        """Call `self.get_features_info()` method to generate features_info list and then generate feature hooks.
        
        If `self.get_features_info()` isn't overrided in an inherited class then the features_info and hooks 
        will be None and no feature hooks will be created.

        Raises:
            ValueError: If overriden `self.get_features_info()` method return not FeatureInfo list.
        """
        features_info = self.get_features_info(*args, **kwargs)
        # Check if all self._features_info is FeatureInfo types.
        self.__check_features_info_types(features_info)
        self.__features_info = features_info
        self.__feature_hooks = FeatureHooks(features_info, self.named_modules())

    @staticmethod
    def __check_features_info_types(features_info: Any):
        """Check if all self._features_info is FeatureInfo class.

        Args:
            features_info: Hooks feature info list that must be checked.

        Raises:
            ValueError: If features_info is not list, or if any value in features_info is not FeatureInfo class.
        """
        if not isinstance(features_info, list):
            raise ValueError('All features_info must be list.')

        for feature in features_info:
            if type(feature) != FeatureInfo:
                raise ValueError('All features_info must be FeatureInfo class.')

    def forward_stage_features(self, *input: Any) -> Tuple[torch.Tensor, List[torch.Tensor]]:
        """Return model output with hooks features.
        
        Returns:
            last_features: Module forward method output.
            hooks_features: Hooks outputs.
        """
        device = list(*input)[0].device
        last_features = self.forward(*input)
        if self.__feature_hooks is not None:
            hooks_features = self.__feature_hooks.get_features(device)
            hooks_features = list(hooks_features.values())
        else:
            hooks_features = None
        return last_features, hooks_features

    def _get_hooks_output_channels(self) -> List[int]:
        """Generate hooks output channels numbers.
        
        Returns:
            output_hooks_channels: Hooks output channels numbers.
        """
        if self.__features_info is None:
            output_hooks_channels = None
        else:
            output_hooks_channels = [feature.num_channels for feature in self.__features_info]
        return output_hooks_channels

    @abstractmethod
    def get_forward_output_channels(self) -> Union[int, List[int]]:
        """Set output channels for Module forward pass.
        
        Returns: Outpus channels.
        """
        pass
    
    def get_output_channels(self) -> Tuple[Union[int, List[int]], List[int]]:
        """Create forward an hooks channels numbers.
        
        Returns:
            forward_channels: Forward pass output channels.
            hooks_channels: Hooks output channels.
        """
        forward_channels = self.get_forward_output_channels()
        hooks_channels = self._get_hooks_output_channels()
        return forward_channels, hooks_channels

    @property
    def features_info(self):
        return self.__features_info

    @property
    def feature_hooks(self):
        return self.__feature_hooks



from torchok.models.heads.base import AbstractHead
from torchok.models.heads import classification
from torchok.models.heads import detection
from torchok.models.heads import representation
from torchok.models.heads import segmentation


__all__ = [
    'AbstractHead',
]



from torchok.tasks.base import BaseTask
from torchok.tasks.classification import ClassificationTask
# from src.tasks.moby import 
# from src.tasks.segmentation import 
# from src.tasks.representation import 



# from . import tasks
from torchok.tasks import (
    BaseTask,
    ClassificationTask,
)
from . import optim
from . import models
from . import metrics
from . import data
from . import constructor
from . import losses


__all__ = [
    'BaseTask',
    'ClassificationTask',
]



from torch import nn, Tensor

from torchok.models.base import BaseModel


class Identity(BaseModel):
    def __init__(self, in_features):
        super().__init__()
        self.identity = nn.Identity()
        self.out_features = in_features

    def forward(self, x: Tensor, target: Tensor = None) -> Tensor:
        x = self.identity(x)
        return x

    def get_forward_output_channels(self):
        return self.out_features
